/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     block64.c                                                  *
 * Created:       2002-08-24 by Hampa Hug <hhug@student.ethz.ch>             *
 * Last modified: 2002-08-27 by Hampa Hug <hhug@student.ethz.ch>             *
 * Copyright:     (C) 2002 by Hampa Hug <hhug@student.ethz.ch>               *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by  the Free Software Foundation.                                         *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: block64.c 384 2005-12-29 11:30:57Z hampa $ */


#include "block64.h"


void blk64_init (block64_t *blk)
{
	blk->word32[0] = 0;
	blk->word32[1] = 0;
}


void blk64_get_buf (block64_t *blk, const unsigned char buf[8])
{
	unsigned long v;

	v = (unsigned long) buf[0] << 24;
	v |= (unsigned long) buf[1] << 16;
	v |= (unsigned long) buf[2] << 8;
	v |= (unsigned long) buf[3];
	blk->word32[0] = v;

	v = (unsigned long) buf[4] << 24;
	v |= (unsigned long) buf[5] << 16;
	v |= (unsigned long) buf[6] << 8;
	v |= (unsigned long) buf[7];
	blk->word32[1] = v;
}

void blk64_set_buf (const block64_t *blk, unsigned char buf[8])
{
	unsigned long v;

	v = blk->word32[0];
	buf[0] = (unsigned char) ((v >> 24) & 0xff);
	buf[1] = (unsigned char) ((v >> 16) & 0xff);
	buf[2] = (unsigned char) ((v >> 8) & 0xff);
	buf[3] = (unsigned char) (v & 0xff);

	v = blk->word32[1];
	buf[4] = (unsigned char) ((v >> 24) & 0xff);
	buf[5] = (unsigned char) ((v >> 16) & 0xff);
	buf[6] = (unsigned char) ((v >> 8) & 0xff);
	buf[7] = (unsigned char) (v & 0xff);
}

void blk64_set_32 (block64_t *blk, unsigned i, unsigned long val)
{
	if (i < 2) {
		blk->word32[i] = val;
	}
}

void blk64_set_8 (block64_t *blk, unsigned i, unsigned char val)
{
	unsigned j, k;

	j = i / 4;
	k = 8 * (3 - (i & 3));

	blk->word32[j] &= ~(255UL << k);
	blk->word32[j] |= (unsigned long) (val &0xff) << k;
}

void blk64_xor (block64_t *dst, const block64_t *src)
{
	dst->word32[0] ^= src->word32[0];
	dst->word32[1] ^= src->word32[1];
}
