/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     hamming.c                                                  *
 * Created:       2003-06-21 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2003-06-21 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2003 by Hampa Hug <hampa@hampa.ch>                     *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by  the Free Software Foundation.                                         *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: hamming.c 384 2005-12-29 11:30:57Z hampa $ */


#include "hamming.h"


void hamming_calc (unsigned data_n, const unsigned char *data,
	unsigned parity_n, unsigned char *parity)
{
	unsigned      i, j, k;
	unsigned char msk;
	unsigned long bit;

	for (k = 0; k < parity_n; k++) {
		parity[k] = 0;
	}

	msk = 1;

	for (i = 0; i < 8; i++) {
		bit = 1;

		for (j = 0; j < data_n; j++) {
			bit += 1;

			while ((bit & (bit - 1)) == 0) {
				bit += 1;
			}

			if (data[j] & msk) {
				unsigned long t;

				t = bit;
				k = 0;
				while ((k < parity_n) && (t != 0)) {
					if (t & 1) {
						parity[k] ^= msk;
					}

					k += 1;
					t = t >> 1;
				}
			}
		}

		msk = msk << 1;
	}
}

unsigned hamming_correct (unsigned data_n, unsigned char *data,
	unsigned parity_n, const unsigned char *parity1, const unsigned char *parity2)
{
	unsigned      i, j;
	unsigned char msk;
	unsigned long bit;
	unsigned      ret;

	ret = 0;
	msk = 1;

	j = 0;
	while ((j < parity_n) && ((parity1[j] ^ parity2[j]) == 0)) {
		j += 1;
	}

	if (j >= parity_n) {
		return (0);
	}

	for (i = 0; i < 8; i++) {
		bit = 0;
		for (j = 0; j < parity_n; j++) {
			bit = bit << 1;
			if ((parity1[parity_n - j - 1] ^ parity2[parity_n - j - 1]) & msk) {
				bit |= 1;
			}
		}

		if (bit != 0) {
			ret += 1;

			if ((bit & (bit - 1)) != 0) {
				bit -= 1;
				j = bit;
				while (j > 0) {
					if ((j & (j - 1)) == 0) {
						bit -= 1;
					}

					j -= 1;
				}

				if (bit < data_n) {
					data[bit] ^= msk;
				}
			}
		}

		msk = msk << 1;
	}

	return (ret);
}
