#!/bin/bash

# $Id: index.sh 480 2008-08-08 12:19:07Z hampa $


case "$1" in
"--version")
	echo "index (misc-utils) version %MISC_VERSION_STR (%MISC_CFG_DATE %MISC_CFG_TIME)"
	echo ""
	echo "Copyright (C) 2000-2008 by Hampa Hug <hampa@hampa.ch>"
	exit 0
	;;

"-?" | "--help")
	echo "usage: index [options] [files...]"
	echo "  -a, --add               Add files to the index"
	echo "  -c, --clean             Force rewrite of index"
	echo "  -d, --delete            Delete entries in index"
	echo "  -D, --mode1             Enable mode 1 [no]"
	echo "  -e, --ed2k              Generate ed2k links"
	echo "  -f, --fields string     Select index fields"
	echo "  -i, --info              Print file descriptions"
	echo "  -j, --merge string      Merge an index file"
	echo "  -J, --merge-del string  Merge and delete an index file"
	echo "  -k, --check             Process all files in index"
	echo "  -l, --list              List index contents"
	echo "  -ll, --list-long        List index contents verbosely"
	echo "  -lm, --list-long-md5    List index contents verbosely with MD5"
	echo "  -m, --md5               Generate a MD5SUM file"
	echo "  -M string               Merge a md5sum type file"
	echo "  -n, --rename-file       Rename files based on MD5"
	echo "  -N, --rename-index      Rename files in index based on MD5"
	echo "  -r, --restore           Retore files from index"
	echo "  -R, --recursive         Process directories recursively [no]"
	echo "  -s, --sha1              Generate a SHA1SUM file"
	echo "  -S, --sha256            Generate a SHA256SUM file"
	echo "  -u, --update            Update files [default]"
	echo "  -v, --verify            Verify files"
	echo "  -z, --zero-mtime        Set all mtimes to 2000-01-01 [no]"
	echo "  -@                      Read file names from stdin"
	exit 0
	;;
esac

par_index_file="INDEX"
par_index_base=""
par_index_rela=""

par_cnt=0
par_mod=0
par_desc=""
par_comm=""
par_mode=0
par_cols=80

par_fld="size mtime mode md5 sha256"
par_fld_ul="size mtime mode crc md4 md5 sha1 sha224 sha256"
par_fld_all="size mtime mode crc md4 md5 ed2k sha1 sha224 sha256"
par_clean=0
par_upd=1
par_rec=0
par_errcnt=0

unset par_idx_name
unset par_idx_desc
unset par_idx_comm
unset par_idx_source
unset par_idx_size
unset par_idx_crc
unset par_idx_adler32
unset par_idx_md4
unset par_idx_md5
unset par_idx_ed2k
unset par_idx_rmd160
unset par_idx_sha1
unset par_idx_sha224
unset par_idx_sha256
unset par_idx_sha384
unset par_idx_sha512
unset par_idx_sig
unset par_idx_mtime
unset par_idx_mode


# set par_cols to the number of columns
# idx_set_cols [default]
idx_set_cols ()
{
	local tmp

	par_cols=80

	if test -n "$1" ; then
		par_cols=$1
	fi

	tmp=$(type -p tput)
	if test -n "$tmp" ; then
		tmp=$(tput cols)
		if test -n "$tmp" ; then
			par_cols=$tmp
		fi
	fi
}

# idx_print mode left right eol
idx_print ()
{
	local t1=$2
	local t2=$3
	local eol=$4
	local w1
	local w2

	case "$1" in
	"" | "normal")
		printf "\\x1b[0m"
		;;
	"ok")
		printf "\\x1b[32m"
		;;
	"bad")
		printf "\\x1b[31m"
		;;
	"maybe")
		printf "\\x1b[33m"
		;;
	esac

	if test -z "$eol" ; then
		eol="\\n"
	fi

	w2=${#t2}
	if test $w2 -gt $(($par_cols-8)) ; then
		w2=$(($par_cols-8))
	fi
	w1=$(($par_cols-w2))

	printf "%-${w1}s%s$eol" "${t1:0:w1}" "$t2"

	printf "\\x1b[0m"
}

# set par_index_file
idx_get_dbfile ()
{
	local f

	if test -n "$INDEXDB" ; then
		par_index_file=$INDEXDB
		return
	fi

	for f in INDEX ../INDEX ../../INDEX ../../../INDEX ; do
		if test -f "$f" ; then
			par_index_file=$f
			return
		fi
	done

	par_index_file="INDEX"
}

idx_merge_md5sum ()
{
	local i
	local line
	local chk cnt name

	while read line ; do
		case "$line" in
		"" | "#"*)
			continue
			;;
		esac

		chk=${line%% *}
		chk=$(echo "$chk" | tr "A-Z" "a-z")
		cnt=${#chk}

		name=${line:$(($cnt+2))}

		i=$(idx_find_entry "$name")
		if test $i -ge $par_cnt ; then
			par_idx_name[$i]=$name
		fi

		case "$cnt" in
		32)
			if test -z "${par_idx_md5[$i]}" ; then
				par_idx_md5[$i]=$chk
			fi
			;;
		40)
			if test -z "${par_idx_sha1[$i]}" ; then
				par_idx_sha1[$i]=$chk
			fi
			;;
		56)
			if test -z "${par_idx_sha224[$i]}" ; then
				par_idx_sha224[$i]=$chk
			fi
			;;
		64)
			if test -z "${par_idx_sha256[$i]}" ; then
				par_idx_sha256[$i]=$chk
			fi
			;;
		96)
			if test -z "${par_idx_sha384[$i]}" ; then
				par_idx_sha384[$i]=$chk
			fi
			;;
		128)
			if test -z "${par_idx_sha512[$i]}" ; then
				par_idx_sha512[$i]=$chk
			fi
			;;
		esac

		if test $i -ge $par_cnt ; then
			par_cnt=$((i+1))
		fi
	done

	par_mod=1
}

idx_read_v1_file ()
{
	local key
	local val
	local idx

	while read key val ; do
		case "$val" in
		\"*\")
			val=${val%\"}
			val=${val#\"}
			;;
		esac

		case "$key" in
		"")
			break
			;;
		"["*"]")
			break
			;;
		
		"#"*)
			continue
			;;
		"name:")
			if test "x$par_mode" = "x0" ; then
				val=${val##*/}
			fi
			par_idx_name[$par_cnt]=$val
			;;
		"desc:")
			par_idx_desc[$par_cnt]=$val
			;;
		"comm:")
			par_idx_comm[$par_cnt]=$val
			;;
		"source:")
			par_idx_source[$par_cnt]=$val
			;;
		"size:")
			par_idx_size[$par_cnt]=$val
			;;
		"ad32:")
			par_idx_adler32[$par_cnt]=$val
			;;
		"crc:")
			par_idx_crc[$par_cnt]=$val
			;;
		"md4:")
			par_idx_md4[$par_cnt]=$val
			;;
		"md5:")
			par_idx_md5[$par_cnt]=$val
			;;
		"ed2k:")
			par_idx_ed2k[$par_cnt]=$val
			;;
		"rmd160:")
			par_idx_rmd160[$par_cnt]=$val
			;;
		"sha1:")
			par_idx_sha1[$par_cnt]=$val
			;;
		"sha224:")
			par_idx_sha224[$par_cnt]=$val
			;;
		"sha256:")
			par_idx_sha256[$par_cnt]=$val
			;;
		"sha384:")
			par_idx_sha384[$par_cnt]=$val
			;;
		"sha512:")
			par_idx_sha512[$par_cnt]=$val
			;;
		"sign:")
			par_idx_sig[$par_cnt]=$val
			;;
		"mtime:")
			par_idx_mtime[$par_cnt]=$val
			;;
		"mode:")
			par_idx_mode[$par_cnt]=$val
			;;
		esac
	done

	par_cnt=$((par_cnt+1))

	if test "x$key" = "x[file]" ; then
		return 0
	fi

	return 1
}

idx_read_v1 ()
{
	local key
	local val

	while read key val ; do
		case "$val" in
		\"*\")
			val=${val%\"}
			val=${val#\"}
			;;
		esac

		case "$key" in
		"" | "#"*)
			continue
			;;
		"desc:")
			test -z "$par_desc" && par_desc=$val
			;;
		"comm:")
			test -z "$par_comm" && par_comm=$val
			;;
		"version:")
			par_version=${val%%/*}
			if test "x$par_version" != "x0" ; then
				idx_print bad "- index" " [version]" "\n"
				par_errcnt=$(($par_errcnt+1))
			fi
			val=${val#$par_version}
			val=${val#/}
			par_mode=${val%%/*}
			if test -z "$par_mode" ; then
				par_mode=0
			fi
			;;
		"files:")
			;;
		"entries:")
			;;
		"mtime:")
			test -z "$par_mtime" && par_mtime=$val
			;;
		"[index]")
			;;
		"[file]")
			while idx_read_v1_file ; do
				:
			done
			;;
		esac
	done
}

idx_read ()
{
	local md5
	local cnt
	local mod1
	local mod2
	local name

	while read md5 cnt mod1 mod2 perm name ; do
		case "$md5" in
		"" | "#"*)
			continue
			;;
		"[index]")
			idx_read_v1
			break
			;;
		esac

		if test -z "$mod2" ; then
			name=$mod1
			mod1="????-??-??"
			mod2="??:??:??"
			perm="640"
		elif test -z "$name" ; then
			name=$perm
			perm="640"
		fi

		par_idx_name[$par_cnt]=$name
		par_idx_md5[$par_cnt]=$md5
		par_idx_size[$par_cnt]=$cnt
		par_idx_mtime[$par_cnt]="$mod1 $mod2"
		par_idx_mode[$par_cnt]=$perm
		par_cnt=$((par_cnt+1))
	done

	if test -z "$par_index_base" ; then
		tmp=$(dirname "$par_index_file")
		par_index_base=$(cd "$tmp" && pwd)
	fi

	par_index_rela=$(pwd)
	par_index_rela=${par_index_rela#$par_index_base}
	par_index_rela=${par_index_rela#/}

	case "$par_index_rela" in
	"")
		;;
	*/)
		;;
	*)
		par_index_rela="$par_index_rela/"
		;;
	esac
}

idx_write ()
{
	local i
	local j
	local f1 f2 t

	echo "#-----------------------------------------------------------------------------"
	echo "[index]"
	if test -n "$par_desc" ; then
		printf "desc:\\t\\t\"%s\"\\n" "$par_desc"
	fi
	if test -n "$par_comm" ; then
		printf "comm:\\t\\t\"%s\"\\n" "$par_comm"
	fi
	if test "x$par_mode" = "x0" ; then
		printf "version:\\t%u\\n" 0
	else
		printf "version:\\t%u/%u\\n" 0 $par_mode
	fi
	if test "x$par_clean" = "x0" ; then
		printf "entries:\\t%u\\n" "$par_cnt"
	fi
	echo "#-----------------------------------------------------------------------------"

	f1=""

	j=0
	while test $j -lt $par_cnt ; do
		if test "x$par_clean" = "x0" -o -e "${par_index_base}/${par_idx_name[$j]}" ; then
			printf "%d %08d %s\\n" "$j" $(($par_cnt-$j)) "${par_idx_name[$j]}"
		fi

		j=$(($j+1))
	done |\
	sort -k 3 -k 2 |\
	while read i t f2 ; do
		if test "x$f2" = "x$f1" ; then
			continue
		fi

		f1=$f2

		echo ""
		echo "[file]"
		printf "name:\\t\"%s\"\\n" "${par_idx_name[$i]}"
		if test -n "${par_idx_desc[$i]}" ; then
			printf "desc:\\t\"%s\"\\n" "${par_idx_desc[$i]}"
		fi
		if test -n "${par_idx_comm[$i]}" ; then
			printf "comm:\\t\"%s\"\\n" "${par_idx_comm[$i]}"
		fi
		if test -n "${par_idx_source[$i]}" ; then
			printf "source:\\t\"%s\"\\n" "${par_idx_source[$i]}"
		fi
		if test -n "${par_idx_size[$i]}" ; then
			printf "size:\\t%s\\n" "${par_idx_size[$i]}"
		fi
		if test -n "${par_idx_adler32[$i]}" ; then
			printf "ad32:\\t%s\\n" "${par_idx_adler32[$i]}"
		fi
		if test -n "${par_idx_crc[$i]}" ; then
			printf "crc:\\t%s\\n" "${par_idx_crc[$i]}"
		fi
		if test -n "${par_idx_md4[$i]}" ; then
			printf "md4:\\t%s\\n" "${par_idx_md4[$i]}"
		fi
		if test -n "${par_idx_md5[$i]}" ; then
			case "${par_idx_md5[$i]}" in
			"----"*)
				;;
			*)
				printf "md5:\\t%s\\n" "${par_idx_md5[$i]}"
				;;
			esac
		fi
		if test -n "${par_idx_ed2k[$i]}" ; then
			printf "ed2k:\\t%s\\n" "${par_idx_ed2k[$i]}"
		fi
		if test -n "${par_idx_rmd160[$i]}" ; then
			printf "rmd160:\\t%s\\n" "${par_idx_rmd160[$i]}"
		fi
		if test -n "${par_idx_sha1[$i]}" ; then
			printf "sha1:\\t%s\\n" "${par_idx_sha1[$i]}"
		fi
		if test -n "${par_idx_sha224[$i]}" ; then
			printf "sha224:\\t%s\\n" "${par_idx_sha224[$i]}"
		fi
		if test -n "${par_idx_sha256[$i]}" ; then
			printf "sha256:\\t%s\\n" "${par_idx_sha256[$i]}"
		fi
		if test -n "${par_idx_sha384[$i]}" ; then
			printf "sha384:\\t%s\\n" "${par_idx_sha384[$i]}"
		fi
		if test -n "${par_idx_sha512[$i]}" ; then
			printf "sha512:\\t%s\\n" "${par_idx_sha512[$i]}"
		fi
		if test -n "${par_idx_sig[$i]}" ; then
			printf "sign:\\t%s\\n" "${par_idx_sig[$i]}"
		fi
		if test -n "${par_idx_mtime[$i]}" ; then
			case "${par_idx_mtime[$i]}" in
			"????"*)
				;;
			*)
				printf "mtime:\\t\"%s\"\\n" "${par_idx_mtime[$i]}"
				;;
			esac
		fi
		if test -n "${par_idx_mode[$i]}" ; then
			printf "mode:\\t%s\\n" "${par_idx_mode[$i]}"
		fi
		i=$((i+1))
	done
}

idx_make_desc ()
{
	local i

	i=0
	while test $i -lt $par_cnt ; do
		printf "%s\\t%s\\n" "${par_idx_name[$i]}" "${par_idx_desc[$i]}"
		i=$((i+1))
	done |\
	sort
}

idx_make_md5 ()
{
	local i

	i=0
	while test $i -lt $par_cnt ; do
		if test -n "${par_idx_md5[$i]}" ; then
			echo "${par_idx_md5[$i]} *${par_idx_name[$i]}"
		fi
		i=$((i+1))
	done |\
	sort -k 2
}

idx_make_sha1 ()
{
	local i

	i=0
	while test $i -lt $par_cnt ; do
		if test -n "${par_idx_sha1[$i]}" ; then
			echo "${par_idx_sha1[$i]} *${par_idx_name[$i]}"
		fi
		i=$((i+1))
	done |\
	sort -k 2
}

idx_make_sha256 ()
{
	local i

	i=0
	while test $i -lt $par_cnt ; do
		if test -n "${par_idx_sha256[$i]}" ; then
			echo "${par_idx_sha256[$i]} *${par_idx_name[$i]}"
		fi
		i=$((i+1))
	done |\
	sort -k 2
}

idx_make_ed2k ()
{
	local i

	i=0
	while test $i -lt $par_cnt ; do
		if test -n "${par_idx_ed2k[$i]}" -a -n "${par_idx_size[$i]}" ; then
			echo "ed2k://|file|${par_idx_name[$i]}|${par_idx_size[$i]}|${par_idx_ed2k[$i]}|/"
		fi
		i=$((i+1))
	done |\
	sort -k 2
}

idx_clear_entry ()
{
	local i=$1

	par_idx_name[$i]=""
	par_idx_desc[$i]=""
	par_idx_comm[$i]=""
	par_idx_source[$i]=""
	par_idx_size[$i]=""
	par_idx_adler32[$i]=""
	par_idx_crc[$i]=""
	par_idx_md4[$i]=""
	par_idx_md5[$i]=""
	par_idx_ed2k[$i]=""
	par_idx_rmd160[$i]=""
	par_idx_sha1[$i]=""
	par_idx_sha224[$i]=""
	par_idx_sha256[$i]=""
	par_idx_sha384[$i]=""
	par_idx_sha512[$i]=""
	par_idx_sig[$i]=""
	par_idx_mtime[$i]=""
	par_idx_mode[$i]=""
}

# idx_copy_entry dst src
idx_copy_entry ()
{
	local i=$1
	local j=$2

	par_idx_name[$i]=${par_idx_name[$j]}
	par_idx_desc[$i]=${par_idx_desc[$j]}
	par_idx_comm[$i]=${par_idx_comm[$j]}
	par_idx_source[$i]=${par_idx_source[$j]}
	par_idx_size[$i]=${par_idx_size[$j]}
	par_idx_adler32[$i]=${par_idx_adler32[$j]}
	par_idx_crc[$i]=${par_idx_crc[$j]}
	par_idx_md4[$i]=${par_idx_md4[$j]}
	par_idx_md5[$i]=${par_idx_md5[$j]}
	par_idx_ed2k[$i]=${par_idx_ed2k[$j]}
	par_idx_rmd160[$i]=${par_idx_rmd160[$j]}
	par_idx_sha1[$i]=${par_idx_sha1[$j]}
	par_idx_sha224[$i]=${par_idx_sha224[$j]}
	par_idx_sha256[$i]=${par_idx_sha256[$j]}
	par_idx_sha384[$i]=${par_idx_sha384[$j]}
	par_idx_sha512[$i]=${par_idx_sha512[$j]}
	par_idx_sig[$i]=${par_idx_sig[$j]}
	par_idx_mtime[$i]=${par_idx_mtime[$j]}
	par_idx_mode[$i]=${par_idx_mode[$j]}
}

idx_find_entry ()
{
	local src=$1
	local i

	i=0
	while test $i -lt $par_cnt ; do
		if test "x$src" = "x${par_idx_name[$i]}" ; then
			echo "$i"
			return
		fi
		i=$(($i+1))
	done

	echo "$par_cnt"
}

idx_get_dsc ()
{
	local src=$1
	local dst

	if test "x$par_mode" = "x0" ; then
		dst=${src##*/}
	else
		dst="${par_index_rela}$src"
	fi

	echo "$dst"
}

# idx_mode_to_sym mode
idx_mode_to_sym ()
{
	local m=$1

	m=$((0x$m))
	s="-"

	if test $(($m&0x400)) -ne 0 ; then s="${s}r" ; else s="${s}-" ; fi
	if test $(($m&0x200)) -ne 0 ; then s="${s}w" ; else s="${s}-" ; fi
	if test $(($m&0x100)) -ne 0 ; then s="${s}x" ; else s="${s}-" ; fi

	if test $(($m&0x40)) -ne 0 ; then s="${s}r" ; else s="${s}-" ; fi
	if test $(($m&0x20)) -ne 0 ; then s="${s}w" ; else s="${s}-" ; fi
	if test $(($m&0x10)) -ne 0 ; then s="${s}x" ; else s="${s}-" ; fi

	if test $(($m&0x4)) -ne 0 ; then s="${s}r" ; else s="${s}-" ; fi
	if test $(($m&0x2)) -ne 0 ; then s="${s}w" ; else s="${s}-" ; fi
	if test $(($m&0x1)) -ne 0 ; then s="${s}x" ; else s="${s}-" ; fi

	echo "$s"
}

idx_list ()
{
	local i

	i=0
	while test $i -lt $par_cnt ; do
		echo "${par_idx_name[$i]}"
		i=$((i+1))
	done
}

# idx_list_long [withmd5]
idx_list_long ()
{
	local i
	local mode md5 mtime size
	local withmd5=$1

	i=0
	while test $i -lt $par_cnt ; do
		mode=${par_idx_mode[$i]}
		if test -z "$mode" ; then
			mode="644"
		fi
		mode=$(idx_mode_to_sym "$mode")

		mtime=${par_idx_mtime[$i]}
		if test -z "$mtime" ; then
			mtime="????-??-?? ??:??:??"
		fi

		size=${par_idx_size[$i]}
		if test -z "$size" ; then
			size=0
		fi

		if test "x$withmd5" = "x1" ; then
			md5=${par_idx_md5[$i]}
			if test -z "$md5" ; then
				md5="--------------------------------"
			fi

			printf "%s " "$md5"
		fi

		printf "%s %s %10d %s\\n" "$mode" "$mtime" "$size" "${par_idx_name[$i]}"

		i=$(($i+1))
	done
}

# idx_update upd src
idx_update ()
{
	local upd=$1
	local src=$2
	local dsc
	local i
	local f
	local mdig
	local type
	local par
	local val
	local ret

	dsc=$(idx_get_dsc "$src")

	i=$(idx_find_entry "$dsc")

	ret=""

	if test "x${par_idx_name[$i]}" != "x$dsc" ; then
		par_idx_name[$i]=$dsc
		ret="$ret index"
	fi

	mdig=""

	for f in $par_fld ; do
		case "$f" in
		"desc")
			;;
		"comm")
			;;
		"source")
			;;
		"size")
			idx_print normal "= $src" " [size]" "\r"
			val=$(stat -c "%s" -L "$src")
			if test "x${par_idx_size[$i]}" != "x$val" ; then
				par_idx_size[$i]=$val
				ret="$ret size"
			fi
			;;
		"adler32")
			if test -z "${par_idx_adler32[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig adler32"
			fi
			;;
		"crc")
			if test -z "${par_idx_crc[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig crc"
			fi
			;;
		"md4")
			if test -z "${par_idx_md4[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig md4"
			fi
			;;
		"md5")
			if test -z "${par_idx_md5[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig md5"
			fi
			;;
		"ed2k")
			if test -z "${par_idx_ed2k[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig ed2k"
			fi
			;;
		"rmd160")
			if test -z "${par_idx_rmd160[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig rmd160"
			fi
			;;
		"sha1")
			if test -z "${par_idx_sha1[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig sha1"
			fi
			;;
		"sha224")
			if test -z "${par_idx_sha224[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig sha224"
			fi
			;;
		"sha256")
			if test -z "${par_idx_sha256[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig sha256"
			fi
			;;
		"sha384")
			if test -z "${par_idx_sha384[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig sha384"
			fi
			;;
		"sha512")
			if test -z "${par_idx_sha512[$i]}" -o "x$upd" = "x0" ; then
				mdig="$mdig sha512"
			fi
			;;
		"sig")
			if test -z "${par_idx_sig[$i]}" -o "x$upd" = "x0" ; then
				idx_print normal "= $src" " [sig]" "\r"
				val=$(gpg --detach-sign < "$src" 2> /dev/null | hex -e -c 0 -p 0 -l)
				if test "x${par_idx_sig[$i]}" != "x$val" ; then
					par_idx_sig[$i]=$val
					ret="$ret sig"
				fi
			fi
			;;
		"mtime")
			idx_print normal "= $src" " [mtime]" "\r"
			val=$(TZ=UTC stat -c "%y" -L "$src")
			val=${val%%.*}
			if test "x${par_idx_mtime[$i]}" != "x$val" ; then
				par_idx_mtime[$i]=$val
				ret="$ret mtime"
			fi
			;;
		"mode")
			idx_print normal "= $src" " [mode]" "\r"
			val=$(stat -c "%a" -L "$src")
			if test "x${par_idx_mode[$i]}" != "x$val" ; then
				par_idx_mode[$i]=$val
				ret="$ret mode"
			fi
			;;
		esac
	done

	if test -n "$mdig" ; then
		par=""
		for f in $mdig ; do
			par="$par -m $f"
		done

		idx_print normal "= $src" " [${mdig# }]" "\r"

		mdig=$(mdig -t $par -i "$src")

		for val in $mdig ; do
			type=${val%:*}
			val=${val#*:}

			case "$type" in
			"adler32")
				if test "x${par_idx_adler32[$i]}" != "x$val" ; then
					par_idx_adler32[$i]=$val
					ret="$ret adler32"
				fi
				;;
			"crc")
				if test "x${par_idx_crc[$i]}" != "x$val" ; then
					par_idx_crc[$i]=$val
					ret="$ret crc"
				fi
				;;
			"md4")
				if test "x${par_idx_md4[$i]}" != "x$val" ; then
					par_idx_md4[$i]=$val
					ret="$ret md4"
				fi
				;;
			"md5")
				if test "x${par_idx_md5[$i]}" != "x$val" ; then
					par_idx_md5[$i]=$val
					ret="$ret md5"
				fi
				;;
			"ed2k")
				if test "x${par_idx_ed2k[$i]}" != "x$val" ; then
					par_idx_ed2k[$i]=$val
					ret="$ret ed2k"
				fi
				;;
			"rmd160")
				if test "x${par_idx_rmd160[$i]}" != "x$val" ; then
					par_idx_rmd160[$i]=$val
					ret="$ret rmd160"
				fi
				;;
			"sha1")
				if test "x${par_idx_sha1[$i]}" != "x$val" ; then
					par_idx_sha1[$i]=$val
					ret="$ret sha1"
				fi
				;;
			"sha224")
				if test "x${par_idx_sha224[$i]}" != "x$val" ; then
					par_idx_sha224[$i]=$val
					ret="$ret sha224"
				fi
				;;
			"sha256")
				if test "x${par_idx_sha256[$i]}" != "x$val" ; then
					par_idx_sha256[$i]=$val
					ret="$ret sha256"
				fi
				;;
			"sha384")
				if test "x${par_idx_sha384[$i]}" != "x$val" ; then
					par_idx_sha384[$i]=$val
					ret="$ret sha384"
				fi
				;;
			"sha512")
				if test "x${par_idx_sha512[$i]}" != "x$val" ; then
					par_idx_sha512[$i]=$val
					ret="$ret sha512"
				fi
				;;
			esac
		done
	fi

	if test -n "$ret" ; then
		idx_print ok "+ $src" " [${ret# }]"
		par_mod=1
	else
		idx_print normal "  $src" ""
	fi

	if test $par_cnt -le $i ; then
		par_cnt=$(($i+1))
	fi

	return 0
}

idx_verify ()
{
	local src=$1
	local dsc
	local i
	local f
	local mdig
	local type
	local val
	local par
	local bad
	local vfy

	dsc=$(idx_get_dsc "$src")

	i=$(idx_find_entry "$dsc")

	if test $i -ge $par_cnt ; then
		idx_print maybe "! $src" " [index]"
		par_errcnt=$(($par_errcnt+1))
		return 1
	fi

	bad=""
	vfy=""
	mdig=""

	for f in $par_fld ; do
		case "$f" in
		"desc")
			;;
		"comm")
			;;
		"source")
			;;
		"size")
			if test -n "${par_idx_size[$i]}" ; then
				idx_print normal "= $src" " [size]" "\r"
				val=$(stat -c "%s" -L "$src")
				if test "x${par_idx_size[$i]}" != "x$val" ; then
					bad="$bad size"
				else
					vfy="$vfy size"
				fi
			fi
			;;
		"adler32")
			if test -n "${par_idx_adler32[$i]}" ; then
				mdig="$mdig adler32"
			fi
			;;
		"crc")
			if test -n "${par_idx_crc[$i]}" ; then
				mdig="$mdig crc"
			fi
			;;
		"md4")
			if test -n "${par_idx_md4[$i]}" ; then
				mdig="$mdig md4"
			fi
			;;
		"md5")
			if test -n "${par_idx_md5[$i]}" ; then
				mdig="$mdig md5"
			fi
			;;
		"ed2k")
			if test -n "${par_idx_ed2k[$i]}" ; then
				mdig="$mdig ed2k"
			fi
			;;
		"rmd160")
			if test -n "${par_idx_rmd160[$i]}" ; then
				mdig="$mdig rmd160"
			fi
			;;
		"sha1")
			if test -n "${par_idx_sha1[$i]}" ; then
				mdig="$mdig sha1"
			fi
			;;
		"sha224")
			if test -n "${par_idx_sha224[$i]}" ; then
				mdig="$mdig sha224"
			fi
			;;
		"sha256")
			if test -n "${par_idx_sha256[$i]}" ; then
				mdig="$mdig sha256"
			fi
			;;
		"sha384")
			if test -n "${par_idx_sha384[$i]}" ; then
				mdig="$mdig sha384"
			fi
			;;
		"sha512")
			if test -n "${par_idx_sha512[$i]}" ; then
				mdig="$mdig sha512"
			fi
			;;
		"sig")
			if test -n "${par_idx_sig[$i]}" ; then
				idx_print normal "= $src" " [sig]" "\r"
				echo "${par_idx_sig[$i]}" | hex -d -n 2 | gpg --verify - "$src" 2> /dev/null
				val=$?
				if test "x$val" != "x0" ; then
					bad="$bad sig"
				else
					vfy="$vfy sig"
				fi
			fi
			;;
		"mtime")
			if test -n "${par_idx_mtime[$i]}" ; then
				idx_print normal "= $src" " [mtime]" "\r"
				val=$(TZ=UTC stat -c "%y" -L "$src")
				val=${val%%.*}
				if test "x${par_idx_mtime[$i]}" != "x$val" ; then
					bad="$bad mtime"
				else
					vfy="$vfy mtime"
				fi
			fi
			;;
		"mode")
			if test -n "${par_idx_mode[$i]}" ; then
				idx_print normal "= $src" " [mode]" "\r"
				val=$(stat -c "%a" -L "$src")
				if test "x${par_idx_mode[$i]}" != "x$val" ; then
					bad="$bad mode"
				else
					vfy="$vfy mode"
				fi
			fi
			;;
		esac
	done

	if test -n "$mdig" ; then
		for f in $mdig ; do
			par="$par -m $f"
		done

		idx_print normal "= $src" " [${mdig# }]" "\r"

		mdig=$(mdig -t $par -i "$src")

		for val in $mdig ; do
			type=${val%:*}
			val=${val#*:}

			case "$type" in
			"adler32")
				if test "x${par_idx_adler32[$i]}" != "x$val" ; then
					bad="$bad adler32"
				else
					vfy="$vfy adler32"
				fi
				;;
			"crc")
				if test "x${par_idx_crc[$i]}" != "x$val" ; then
					bad="$bad crc"
				else
					vfy="$vfy crc"
				fi
				;;
			"md4")
				if test "x${par_idx_md4[$i]}" != "x$val" ; then
					bad="$bad md4"
				else
					vfy="$vfy md4"
				fi
				;;
			"md5")
				if test "x${par_idx_md5[$i]}" != "x$val" ; then
					bad="$bad md5"
				else
					vfy="$vfy md5"
				fi
				;;
			"ed2k")
				if test "x${par_idx_ed2k[$i]}" != "x$val" ; then
					bad="$bad ed2k"
				else
					vfy="$vfy ed2k"
				fi
				;;
			"rmd160")
				if test "x${par_idx_rmd160[$i]}" != "x$val" ; then
					bad="$bad rmd160"
				else
					vfy="$vfy rmd160"
				fi
				;;
			"sha1")
				if test "x${par_idx_sha1[$i]}" != "x$val" ; then
					bad="$bad sha1"
				else
					vfy="$vfy sha1"
				fi
				;;
			"sha224")
				if test "x${par_idx_sha224[$i]}" != "x$val" ; then
					bad="$bad sha224"
				else
					vfy="$vfy sha224"
				fi
				;;
			"sha256")
				if test "x${par_idx_sha256[$i]}" != "x$val" ; then
					bad="$bad sha256"
				else
					vfy="$vfy sha256"
				fi
				;;
			"sha384")
				if test "x${par_idx_sha384[$i]}" != "x$val" ; then
					bad="$bad sha384"
				else
					vfy="$vfy sha384"
				fi
				;;
			"sha512")
				if test "x${par_idx_sha512[$i]}" != "x$val" ; then
					bad="$bad sha512"
				else
					vfy="$vfy sha512"
				fi
				;;
			esac
		done
	fi

	if test -n "$bad" ; then
		idx_print bad "- $src" " [${bad# }]"
		par_errcnt=$(($par_errcnt+1))
		return 1
	elif test -n "$vfy" ; then
		idx_print ok "+ $src" " [${vfy# }]"
		return 0
	else
		idx_print normal "  $src" ""
	fi

	return 0
}

# idx_restore file
idx_restore ()
{
	local src=$1
	local dsc
	local i
	local f
	local val
	local val2

	dsc=$(idx_get_dsc "$src")

	i=$(idx_find_entry "$dsc")

	if test $i -ge $par_cnt ; then
		idx_print maybe "! $src" " [index]"
		return
	fi

	rst=""

	for f in $par_fld ; do
		case "$f" in
		"desc")
			;;
		"comm")
			;;
		"source")
			;;
		"size")
			val=$(stat -c "%s" -L "$src")
			if test "x${par_idx_size[$i]}" != "x$val" ; then
				rst="$rst -size"
			fi
			;;
		"adler32")
			;;
		"crc")
			;;
		"md4")
			;;
		"md5")
			;;
		"ed2k")
			;;
		"rmd160")
			;;
		"sha1")
			;;
		"sha224")
			;;
		"sha256")
			;;
		"sha384")
			;;
		"sha512")
			;;
		"sig")
			if test -z "${par_idx_sig[$i]}" ; then
				rst="$rst -sig"
			elif test -e "$src.sig" ; then
				rst="$rst -sig"
			else
				rst="$rst sig"
				echo "${par_idx_sig[$i]}" | hex -d -n 2 > "$src.sig"
			fi
			;;
		"mtime")
			if test -z "${par_idx_mtime[$i]}" ; then
				rst="$rst -mtime"
			else
				val=$(TZ=UTC stat -c "%y" -L "$src")
				val=${val%%.*}
				if test "x${par_idx_mtime[$i]}" != "x$val" ; then
					rst="$rst mtime"
					val=${par_idx_mtime[$i]}
					val=${val// /}
					val=${val//-/}
					val=${val/:/}
					val=${val/:/.}
					TZ=UTC touch -t "$val" "$src"
				fi
			fi
			;;
		"mode")
			if test -z "${par_idx_mode[$i]}" ; then
				rst="$rst -mode"
			else
				val=$(stat -c "%a" -L "$src")
				if test "x${par_idx_mode[$i]}" != "x$val" ; then
					val=${par_idx_mode[$i]}
					chmod "$val" "$src"
					rst="$rst mode"
				fi
			fi
			;;
		esac
	done

	if test -n "$rst" ; then
		idx_print ok "+ $src" " [${rst# }]"
	else
		idx_print normal "  $src" ""
	fi

	return 0
}

# idx_rename inindex file
idx_rename ()
{
	local index=$1
	local src=$2
	local dsc
	local dir
	local md5
	local i

	dsc=$(idx_get_dsc "$src")

	md5=$(md5 -i "$src")

	i=0
	while test $i -lt $par_cnt ; do
		if test "x$md5" = "x${par_idx_md5[$i]}" ; then
			if test "x${par_idx_name[$i]}" != "x$dsc" ; then
				if test "x$index" = "x1" ; then
					idx_print ok "+ $src" " [${par_idx_name[$i]}]"
					par_idx_name[$i]=$dsc
					par_mod=1
				else
					if test -e "${par_idx_name[$i]}" ; then
						idx_print bad "- $src" " [${par_idx_name[$i]}]"
					else
						idx_print ok "+ ${par_idx_name[$i]}" " [$src]"
						dst="$par_index_base/${par_idx_name[$i]}"
						mkdir -p "$(dirname "$dst")"
						mv -f "$src" "$dst"
					fi
				fi
			else
				idx_print normal "  $src" ""
			fi
			return
		fi
		i=$(($i+1))
	done

	echo "! $src"
}

# idx_delete file
idx_delete ()
{
	local src=$1
	local dsc
	local i

	dsc=$(idx_get_dsc "$src")

	i=$(idx_find_entry "$dsc")

	if test $i -ge $par_cnt ; then
		return
	fi

	i=$(($i+1))
	while test $i -lt $par_cnt ; do
		idx_copy_entry $(($i-1)) $i
		i=$(($i+1))
	done

	idx_print ok "+ $src" " [delete]"

	par_cnt=$(($par_cnt-1))

	par_mod=1
}

idx_zero_mtime ()
{
	local i

	i=0
	while test $i -lt $par_cnt ; do
		if test -n "${par_idx_name[$i]}" ; then
			if test "x${par_idx_mtime[$i]}" != "x2000-01-01 00:00:00" ; then
				idx_print ok "+ ${par_idx_name[$i]}" " [mtime]"
				par_idx_mtime[$i]="2000-01-01 00:00:00"
				par_mod=1
			fi
		fi

		i=$(($i+1))
	done

}

# process_file action file
process_file ()
{
	local act=$1
	local src=$2
	local f

	src=${src#./}
	src=${src%/}

	case "$src" in
	"INDEX" | */"INDEX")
		echo "  $src"
		return 0
		;;
	esac

	if test -d "$src" ; then
		if test "x$par_rec" != "x1" ; then
			return 0
		fi
		for f in "$src"/* ; do
			if test -e "$f" ; then
				process_file "$act" "$f"
			fi
		done
		return 0
	fi

	if test ! -f "$src" ; then
		if test "x$act" != "xd" ; then
			idx_print bad "- $src" " [no file]"
			return 1
		fi
	fi

	case "$act" in
	"a")
		idx_update 0 "$src"
		;;
	"d")
		idx_delete "$src"
		;;
	"n")
		idx_rename 0 "$src"
		;;
	"N")
		idx_rename 1 "$src"
		;;
	"r")
		idx_restore "$src"
		;;
	"u")
		idx_update 1 "$src"
		;;
	"v")
		idx_verify "$src"
		;;
	*)
		return 1
		;;
	esac
}

# idx_process action
idx_process ()
{
	local act=$1
	local i

	i=0
	while test $i -lt $par_cnt ; do
		process_file "$act" "${par_idx_name[$i]}"
		i=$((i+1))
	done
}


idx_get_dbfile

if test ! -e "$par_index_file" ; then
	touch "$par_index_file"
fi

idx_read < "$par_index_file"

par_mod=0
par_ret=0

act="u"

idx_set_cols

while test $# -gt 0 ; do
	case "$1" in
	"-a" | "--add")
		act="a"
		;;
	"-aa")
		act="a"
		par_fld=$par_fld_all
		;;
	"-aaa")
		act="a"
		par_fld="$par_fld_all sig"
		;;
	"-c" | "--clean")
		par_mod=1
		;;
	"-C")
		par_mod=1
		par_clean=1
		;;
	"-d" | "--delete")
		act="d"
		;;
	"-D" | "--mode1")
		par_mode=1
		;;
	"-e" | "--ed2k")
		idx_make_ed2k
		;;
	"-f" | "--fields")
		case "$2" in
		"all")
			par_fld=$par_fld_all
			;;
		"chk")
			par_fld="crc md4 md5 rmd160 sha1 sha256"
			;;
		"+"*)
			par_fld="$par_fld ${2#+}"
			;;
		"-"*)
			for f in ${2#-} ; do
				par_fld=${par_fld/$f/}
			done
			;;
		*)
			par_fld=$2
			;;
		esac
		shift 1
		;;
	"-i" | "--info")
		idx_make_desc
		;;
	"-j" | "--merge")
		idx_read < "$2"
		par_mod=1
		shift 1
		;;
	"-J" | "--merge-del")
		idx_read < "$2"
		rm -f "$2"
		par_mod=1
		shift 1
		;;
	"-k" | "--check")
		idx_process "$act"
		;;
	"-l" | "--list")
		idx_list
		;;
	"-ll" | "--list-long")
		idx_list_long 0
		;;
	"-lm" | "--list-long-md5")
		idx_list_long 1
		;;
	"-m" | "--md5")
		idx_make_md5
		;;
	"-M")
		if test -f "$2" ; then
			idx_merge_md5sum < "$2"
		fi
		shift 1
		;;
	"--mode")
		par_mode=$2
		shift 1
		;;
	"-n" | "--rename-file")
		act="n"
		;;
	"-N" | "--rename-index")
		act="N"
		;;
	"-r" | "--restore")
		act="r"
		;;
	"-R" | "--recursive")
		par_rec=1
		;;
	"-s" | "--sha1")
		idx_make_sha1
		;;
	"-S" | "--sha256")
		idx_make_sha256
		;;
	"-u" | "--update")
		act="u"
		;;
	"-uu")
		act="u"
		par_fld=$par_fld_all
		;;
	"-ul")
		act="u"
		par_fld=$par_fld_ul
		;;
	"-uuu")
		act="u"
		par_fld="$par_fld_all sig"
		;;
	"-v" | "--verify")
		act="v"
		;;
	"-vv")
		act="v"
		par_fld=$par_fld_all
		;;
	"-vvv")
		act="v"
		par_fld="$par_fld_all sig"
		;;
	"-z" | "--zero-mtime")
		idx_zero_mtime
		;;
	"-@")
		while read fname ; do
			if ! process_file "$act" "$fname" ; then
				par_ret=1
			fi
		done
		;;
	-*)
		echo "$0: unknown option ($1)" >&2
		exit 1
		;;
	"INDEX" | "$db")
		echo "  $1"
		;;
	*)
		if ! process_file "$act" "$1" ; then
			par_ret=1
		fi
		;;
	esac
	shift 1
done

if test "x$par_mod" = "x1" ; then
	idx_write > "$par_index_file"
fi

if test $par_errcnt -ne 0 ; then
	if test $par_errcnt -eq 1 ; then
		idx_print bad "- $par_errcnt error"
	else
		idx_print bad "- $par_errcnt errors"
	fi
fi

exit $par_ret
