/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     sha1.h                                                     *
 * Created:       2004-04-10 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2004-04-10 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2004 Hampa Hug <hampa@hampa.ch>                        *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: sha1.h 384 2005-12-29 11:30:57Z hampa $ */


#ifndef MISC_SHA1_H
#define MISC_SHA1_H 1


/*!***************************************************************************
 * @short The SHA1 context
 *****************************************************************************/
typedef struct {
	unsigned      buf_i;
	unsigned char buf[64];

	unsigned long A;
	unsigned long B;
	unsigned long C;
	unsigned long D;
	unsigned long E;

	unsigned long cnt[2];
} sha1_t;



/*!***************************************************************************
 * @short Initialize an SHA1 context
 *****************************************************************************/
void sha1_init (sha1_t *sha1);

/*!***************************************************************************
 * @short Copy an SHA1 context
 *****************************************************************************/
void sha1_copy (sha1_t *dst, const sha1_t *src);

/*!***************************************************************************
 * @short Calculate an SHA1 sum
 * @param sha1 The SHA1 context
 * @param buf  The data
 * @param cnt  The number of bytes at buf
 *
 * This function can be called multiple times to add more bytes.
 *****************************************************************************/
void sha1_calc (sha1_t *sha1, const void *buf, unsigned long cnt);

/*!***************************************************************************
 * @short End calculating an SHA1 sum
 * @param sha1 The SHA1 context
 *
 * This function must be called after the last call to sha1_calc() and
 * before getting the sha1 sum.
 *****************************************************************************/
void sha1_done (sha1_t *sha1);

/*!***************************************************************************
 * @short Get a 32bit word of the SHA1 sum
 * @param sha1 The SHA1 context
 * @param i    The word index. 0 <= i <= 4.
 *****************************************************************************/
unsigned long sha1_get_sum_i (sha1_t *sha1, unsigned i);

/*!***************************************************************************
 * @short Copy the SHA1 sum to a byte buffer
 *
 * The byte buffer must be at least 20 bytes.
 *****************************************************************************/
void sha1_get_sum (sha1_t *sha1, void *buf);


#endif
