/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     pipe.c                                                     *
 * Created:       2000-08-12 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-07-10 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2000-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: pipe.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/common.h>
#include <lib/strarg.h>

#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/poll.h>

#include "pipe.h"


static const char *argv0 = NULL;


static
void prt_help (void)
{
	fputs ("usage: pipe sep cmd {sep cmd} {sep sep cmd {sep cmd}}\n", stderr);
	fflush (stderr);
}

static
pipe_arg_t *arg_new (char *str)
{
	pipe_arg_t *arg;

	arg = (pipe_arg_t *) malloc (sizeof (pipe_arg_t));
	if (arg == NULL) {
		return (NULL);
	}

	arg->next = NULL;

	arg->arg = (char *) malloc (strlen (str) + 1);
	if (arg->arg == NULL) {
		free (arg);
		return (NULL);
	}

	strcpy (arg->arg, str);

	return (arg);
}

#if 0
static
void arg_del (pipe_arg_t *arg)
{
	if (arg != NULL) {
		free (arg->arg);
		free (arg);
	}
}
#endif

static
pipe_cmd_t *cmd_new (char *str)
{
	pipe_cmd_t *cmd;

	cmd = (pipe_cmd_t *) malloc (sizeof (pipe_cmd_t));
	if (cmd == NULL) {
		return (NULL);
	}

	cmd->next = NULL;
	cmd->arg = NULL;

	cmd->cmd = (char *) malloc (strlen (str) + 1);
	if (cmd->cmd == NULL) {
		free (cmd);
		return (NULL);
	}

	strcpy (cmd->cmd, str);

	return (cmd);
}

#if 0
static
void cmd_del (pipe_cmd_t *cmd)
{
	pipe_arg_t *arg1, *arg2;

	if (cmd == NULL) {
		return;
	}

	arg1 = cmd->arg;
	while (arg1 != NULL) {
		arg2 = arg1->next;
		arg1->next = NULL;
		arg_del (arg1);
		arg1 = arg2;
	}

	free (cmd->cmd);
	free (cmd);
}
#endif

static
void cmd_arg_add (pipe_cmd_t *cmd, pipe_arg_t *arg)
{
	pipe_arg_t *tmp;

	if ((cmd == NULL) || (arg == NULL)) {
		return;
	}

	if (cmd->arg == NULL) {
		cmd->arg = arg;
	}
	else {
		tmp = cmd->arg;
		while (tmp->next != NULL) {
			tmp = tmp->next;
		}

		tmp->next = arg;
	}
}

static
unsigned cmd_arg_get_cnt (pipe_cmd_t *cmd)
{
	unsigned   n;
	pipe_arg_t *arg;

	n = 0;
	arg = cmd->arg;

	while (arg != NULL) {
		arg = arg->next;
		n++;
	}

	return (n);
}

static
int cmd_run (pipe_cmd_t *cmd, int fd_inp, int fd_out)
{
	unsigned   i, n;
	pipe_arg_t *arg;
	char       **argv;

	n = cmd_arg_get_cnt (cmd);

	argv = (char **) malloc ((n + 2) * sizeof (char *));
	if (argv == NULL) {
		return (1);
	}

	argv[0] = cmd->cmd;
	argv[n + 1] = NULL;

	arg = cmd->arg;
	for (i = 1; i <= n; i++) {
		argv[i] = arg->arg;
		arg = arg->next;
	}

	if (fd_inp != 0) {
		dup2 (fd_inp, 0);
		close (fd_inp);
	}

	if (fd_out != 1) {
		dup2 (fd_out, 1);
		close (fd_out);
	}

	execvp (cmd->cmd, argv);

	fprintf (stderr, "%s: exec failed (%s)\n", argv0, cmd->cmd);
	exit (1);

	return (1);
}

static
pipe_t *pipe_new (void)
{
	pipe_t *p;

	p = (pipe_t *) malloc (sizeof (pipe_t));
	if (p == NULL) {
		return (NULL);
	}

	p->next = NULL;
	p->cmd = NULL;

	return (p);
}

#if 0
static
void pipe_del (pipe_t *p)
{
	pipe_cmd_t *cmd1, *cmd2;

	if (p == NULL) {
		return;
	}

	cmd1 = p->cmd;
	while (cmd1 != NULL) {
		cmd2 = cmd1->next;
		cmd1->next = NULL;
		cmd_del (cmd1);
		cmd1 = cmd2;
	}

	free (p);
}
#endif

static
void pipe_cmd_add (pipe_t *p, pipe_cmd_t *cmd)
{
	pipe_cmd_t *tmp;

	if ((p == NULL) || (cmd == NULL)) {
		return;
	}

	if (p->cmd == NULL) {
		p->cmd = cmd;
	}
	else {
		tmp = p->cmd;
		while (tmp->next != NULL) {
			tmp = tmp->next;
		}

		tmp->next = cmd;
	}
}

static
int pipe_run (pipe_cmd_t *cmd, int fd_inp, int fd_out)
{
	int r;
	int fd[2];

	if (cmd == NULL) {
		return (1);
	}

	while (cmd->next != NULL) {
		r = pipe (fd);
		if (r != 0) {
			fprintf (stderr, "%s: creating pipe failed\n", argv0);
			return (1);
		}

		switch (fork()) {
			case -1:
				fprintf (stderr, "%s: fork failed\n", argv0);
				return (1);

			case 0:
				close (fd[1]);
				close (fd_inp);
				fd_inp = fd[0];
				break;

			default:
				close (fd[0]);
				close (fd_out);
				cmd_run (cmd, fd_inp, fd[1]);
				return (1);
		}

		cmd = cmd->next;
	}

	cmd_run (cmd, fd_inp, fd_out);

	return (1);
}

static
void plst_add_pipe (pipe_t **pl, pipe_t *p)
{
	pipe_t *tmp;

	if (*pl == NULL) {
		*pl = p;
	}
	else {
		tmp = *pl;
		while (tmp->next != NULL) {
			tmp = tmp->next;
		}

		tmp->next = p;
	}
}

static
int plst_run (pipe_t *p, int fd_inp, int fd_out)
{
	int    r;
	int    fd1[2];
	int    fd2[2];
	pipe_t *tmp1, *tmp2;

	if (p == NULL) {
		return (1);
	}

	tmp1 = p;
	while (tmp1 != NULL) {
		r = pipe (fd1);
		r |= pipe (fd2);

		if (r != 0) {
			fprintf (stderr, "%s: creating pipe failed\n", argv0);
			return (1);
		}

		switch (fork()) {
			case -1:
				fprintf (stderr, "%s: fork failed\n", argv0);
				return (1);

			case 0:
				tmp2 = p;
				while (tmp2 != tmp1) {
					close (tmp2->fd_inp);
					close (tmp2->fd_out);
					tmp2 = tmp2->next;
				}

				close (fd1[1]);
				close (fd2[0]);

				pipe_run (tmp1->cmd, fd1[0], fd2[1]);

				return (1);

			default:
				tmp1->fd_inp = fd1[1];
				tmp1->fd_out = fd2[0];

				close (fd1[0]);
				close (fd2[1]);

				break;
		}

		tmp1 = tmp1->next;
	}

	return (0);
}

static
int plst_drive_inp (pipe_t *p, int fd_inp)
{
	unsigned      cnt;
	unsigned char buf[PIPE_BUF_SIZE];
	pipe_t        *tmp;

	tmp = p;
	while (tmp != NULL) {
		close (tmp->fd_out);
		tmp = tmp->next;
	}

	cnt = (unsigned) read (fd_inp, buf, PIPE_BUF_SIZE);

	while (cnt > 0) {
		tmp = p;
		while (tmp != NULL) {
			write (tmp->fd_inp, buf, cnt);
			tmp = tmp->next;
		}

		cnt = (unsigned) read (fd_inp, buf, PIPE_BUF_SIZE);
	}

	tmp = p;
	while (tmp != NULL) {
		close (tmp->fd_inp);
		tmp = tmp->next;
	}

	return (0);
}

static
int plst_drive_out (pipe_t *p, int fd_out)
{
	int           r;
	unsigned      i, j, n;
	struct pollfd *pfd;
	pipe_t        *tmp;
	unsigned char buf[PIPE_BUF_SIZE];
	unsigned      buf_n;

	n = 0;
	tmp = p;
	while (tmp != NULL) {
		close (tmp->fd_inp);

		n += 1;
		tmp = tmp->next;
	}

	pfd = (struct pollfd *) malloc (n * sizeof (struct pollfd));
	if (pfd == NULL) {
		return (1);
	}

	tmp = p;
	for (i = 0; i < n; i++) {
		pfd[i].fd = tmp->fd_out;
		pfd[i].events = POLLIN;
		tmp = tmp->next;
	}

	while (n > 0) {
		r = poll (pfd, n, -1);

		i = 0;
		while (i < n) {
			if (pfd[i].revents & POLLIN) {
				buf_n = (unsigned) read (pfd[i].fd, buf, PIPE_BUF_SIZE);
				write (fd_out, buf, buf_n);

				i += 1;
			}
			else if (pfd[i].revents != 0) {
				for (j = i + 1; j < n; j++) {
					pfd[j - 1] = pfd[j];
				}

				n -= 1;
			}
			else {
				i += 1;
			}
		}
	}

	return (0);
}

static
int plst_drive (pipe_t *p, int fd_inp, int fd_out)
{
	switch (fork()) {
		case -1:
			fprintf (stderr, "%s: fork failed\n", argv0);
			return (1);

		case 0:
			plst_drive_out (p, fd_out);
			exit (0);
			return (0);

		default:
			plst_drive_inp (p, fd_inp);
			return (0);
	}
}

int main (int argc, char *argv[])
{
	int        r;
	int        i;
	char       *sep;
	pipe_t     *lst;
	pipe_t     *cur_pipe;
	pipe_cmd_t *cur_cmd;

	argv0 = argv[0];

	if (argc < 2) {
		prt_help();
		return (1);
	}

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("pipe");
			return (EXIT_SUCCESS);
		}
	}

	sep = argv[1];
	if (strlen (sep) < 1) {
		fprintf (stderr, "%s: bad separator (%s)\n", argv0, argv[1]);
		return (1);
	}

	lst = NULL;
	cur_pipe = NULL;
	cur_cmd = NULL;

	i = 2;
	while (i < argc) {
		if (strcmp (argv[i], sep) == 0) {
			if (cur_cmd == NULL) {
				cur_pipe = NULL;
			}
			else {
				cur_cmd = NULL;
			}
		}
		else {
			if (cur_cmd == NULL) {
				cur_cmd = cmd_new (argv[i]);

				if (cur_pipe == NULL) {
					cur_pipe = pipe_new();
					plst_add_pipe (&lst, cur_pipe);
				}

				pipe_cmd_add (cur_pipe, cur_cmd);
			}
			else {
				cmd_arg_add (cur_cmd, arg_new (argv[i]));
			}
		}

		i += 1;
	}

	r = plst_run (lst, 0, 1);
	if (r != 0) {
		return (EXIT_FAILURE);
	}

	r = plst_drive (lst, 0, 1);
	if (r != 0) {
		return (EXIT_FAILURE);
	}

	while (lst != NULL) {
		wait (NULL);
		lst = lst->next;
	}

	return (EXIT_SUCCESS);
}
