/*****************************************************************************
 * misc-utils                                                                *
 *****************************************************************************/

/*****************************************************************************
 * File name:     readlink.c                                                 *
 * Created:       2003-04-07 by Hampa Hug <hampa@hampa.ch>                   *
 * Last modified: 2005-07-10 by Hampa Hug <hampa@hampa.ch>                   *
 * Copyright:     (C) 2002-2005 Hampa Hug <hampa@hampa.ch>                   *
 *****************************************************************************/

/*****************************************************************************
 * This program is free software. You can redistribute it and / or modify it *
 * under the terms of the GNU General Public License version 2 as  published *
 * by the Free Software Foundation.                                          *
 *                                                                           *
 * This program is distributed in the hope  that  it  will  be  useful,  but *
 * WITHOUT  ANY   WARRANTY,   without   even   the   implied   warranty   of *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU  General *
 * Public License for more details.                                          *
 *****************************************************************************/

/* $Id: readlink.c 384 2005-12-29 11:30:57Z hampa $ */


#include <lib/common.h>
#include <lib/strarg.h>

#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>

#include "readlink.h"

#ifndef PATH_MAX
#define PATH_MAX 256
#endif


static const char *argv0 = NULL;

static int par_nl = 1;
static int par_realpath = 0;


static
void prt_help (void)
{
	fputs (
		"usage: readlink [options] file\n"
		"  -n, --no-newline    Don't print a newline after the link\n"
		"  -f, --canonicalize  Follow every element of a path\n",
		stdout
	);

	fflush (stdout);
}

static
int do_readlink (const char *name)
{
	int  n;
	char buf[PATH_MAX];

	n = readlink (name, buf, PATH_MAX);
	if (n < 0) {
		return (1);
	}

	fwrite (buf, 1, n, stdout);

	if (par_nl) {
		fputs ("\n", stdout);
	}

	fflush (stdout);

	return (0);
}

static
int do_realpath (const char *name)
{
	char *ret;
	char buf[PATH_MAX];

	ret = realpath (name, buf);

	if (ret == NULL) {
		return (1);
	}

	fputs (ret, stdout);

	if (par_nl) {
		fputs ("\n", stdout);
	}

	fflush (stdout);

	return (0);
}

int main (int argc, char *argv[])
{
	int i;

	argv0 = argv[0];

	if (argc == 2) {
		if (str_isarg2 (argv[1], "?", "help")) {
			prt_help();
			return (EXIT_SUCCESS);
		}
		else if (str_isarg2 (argv[1], NULL, "version")) {
			prt_version ("readlink");
			return (EXIT_SUCCESS);
		}
	}

	i = 1;
	while (i < argc) {
		if (str_isarg2 (argv[i], "n", "no-newline")) {
			par_nl = 0;
		}
		else if (str_isarg2 (argv[i], "f", "canonicalize")) {
			par_realpath = 1;
		}
		else {
			if (par_realpath) {
				if (do_realpath (argv[i])) {
					return (EXIT_FAILURE);
				}
			}
			else {
				if (do_readlink (argv[i])) {
					return (EXIT_FAILURE);
				}
			}
		}

		i += 1;
	}

	return (EXIT_SUCCESS);
}
